/***********************************************************************
 *                                                                     *
 * $Id: hpgspclbitmapfont.c 622 2011-05-23 08:14:15Z hpgs $
 *                                                                     *
 * hpgs - HPGl Script, a hpgl/2 interpreter, which uses a Postscript   *
 *        API for rendering a scene and thus renders to a variety of   *
 *        devices and fileformats.                                     *
 *                                                                     *
 * (C) 2004-2009 ev-i Informationstechnologie GmbH  http://www.ev-i.at *
 *                                                                     *
 * Author: Wolfgang Glas                                               *
 *                                                                     *
 *  hpgs is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public          *
 * License as published by the Free Software Foundation; either        *
 * version 2.1 of the License, or (at your option) any later version.  *
 *                                                                     *
 * hpgs is distributed in the hope that it will be useful,             *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   *
 * Lesser General Public License for more details.                     *
 *                                                                     *
 * You should have received a copy of the GNU Lesser General Public    *
 * License along with this library; if not, write to the               *
 * Free Software  Foundation, Inc., 59 Temple Place, Suite 330,        *
 * Boston, MA  02111-1307  USA                                         *
 *                                                                     *
 ***********************************************************************
 *                                                                     *
 * The implementation of the PCL bitmap font interpreter.              *
 *                                                                     *
 ***********************************************************************/

#include <hpgsreader.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#if defined ( __MINGW32__ ) || defined ( _MSC_VER )
#include<malloc.h>
#else
#include<alloca.h>
#endif

// #define DEBUG_PCL_CHARACTERS

static int pcl_read_all(hpgs_reader *reader, unsigned char *buf, size_t buf_len)
{
    size_t n,nread=0;

    while (!reader->interrupted && nread < buf_len &&
        (n=hpgs_istream_read(buf+nread,1,buf_len-nread,reader->in)) > 0) {

        nread += n;
    }

    if (nread < buf_len) return -1;

    return 0;
}

void hpgs_reader_pcl_bitmap_font_destroy (hpgs_reader_pcl_bitmap_font *font) {

    int i;
    for (i=0;i<256;++i) {

        if (font->characters[i])
            hpgs_reader_pcl_bitmap_character_destroy(font->characters[i]);
    }
    free(font);
}

void hpgs_reader_pcl_bitmap_character_destroy (hpgs_reader_pcl_bitmap_character *character) {

    if (character->image)
        hpgs_image_destroy(character->image);

    free(character);
}

static size_t hpgs_reader_pcl_bitmap_font_lower_bound(hpgs_reader *reader,
        int font_id) {

    size_t i0 = 0;
    size_t i1 = reader->pcl_bitmap_fonts_size;

    while (i1 > i0) {

        size_t i = i0+(i1-i0)/2;

        if (reader->pcl_bitmap_fonts[i]->font_id < font_id)
          i0 = i+1;
        else
          i1 = i;
    }

    return i0;
}

static hpgs_reader_pcl_bitmap_font *hpgs_reader_pcl_bitmap_font_new (
        hpgs_reader *reader,
        int font_id,double pitch, double height,
        double baseline_position,int x_resolution,int y_resolution) {

    hpgs_reader_pcl_bitmap_font *ret = (hpgs_reader_pcl_bitmap_font*)malloc(sizeof(hpgs_reader_pcl_bitmap_font));

    if (!ret) return NULL;

    memset(ret,0,sizeof(hpgs_reader_pcl_bitmap_font));

    ret->font_id = font_id;
    ret->pitch   = pitch;
    ret->height  = height;
    ret->baseline_position = baseline_position;
    ret->x_resolution = x_resolution;
    ret->y_resolution = y_resolution;

    size_t  i = hpgs_reader_pcl_bitmap_font_lower_bound(reader,font_id);

    if (i < reader->pcl_bitmap_fonts_size && reader->pcl_bitmap_fonts[i]->font_id == font_id) {

        hpgs_reader_pcl_bitmap_font *old_font = reader->pcl_bitmap_fonts[i];
        hpgs_reader_pcl_bitmap_font_destroy(old_font);

        if (reader->pcl_current_bitmap_font == old_font)
            reader->pcl_current_bitmap_font = ret;

        if (reader->pcl_primary_bitmap_font == old_font)
            reader->pcl_primary_bitmap_font = ret;
    }
    else {

        if (HPGS_GROW_ARRAY_MIN_SIZE(reader,hpgs_reader_pcl_bitmap_font,pcl_bitmap_fonts,pcl_bitmap_fonts_size,pcl_bitmap_fonts_alloc_size,16)) {

            free(ret);
            return NULL;
        }

        if (i < reader->pcl_bitmap_fonts_size)
            memmove(reader->pcl_bitmap_fonts+(i+1),reader->pcl_bitmap_fonts+i,
                    (reader->pcl_bitmap_fonts_size-i)*sizeof(hpgs_reader_pcl_bitmap_font*));

        ++reader->pcl_bitmap_fonts_size;
    }
    reader->pcl_bitmap_fonts[i] = ret;

    return ret;
}

int hpgs_reader_pcl_cmd_font_id              (hpgs_reader *reader, int arg) {

    size_t i = hpgs_reader_pcl_bitmap_font_lower_bound(reader,arg);

    if (i < reader->pcl_bitmap_fonts_size &&
            reader->pcl_bitmap_fonts[i]->font_id == arg) {
        reader->pcl_current_bitmap_font = reader->pcl_bitmap_fonts[i];
    }
    else {
        reader->pcl_current_bitmap_font = NULL;
    }
    reader->pcl_current_bitmap_font_id = arg;

    return 0;
}

int hpgs_reader_pcl_cmd_font_header(hpgs_reader *reader, int arg) {

    if (arg != 70) {

        hpgs_log(hpgs_i18n("PCL Font Header command with invalid length %d.\n"),arg);
        return 0;
    }

    unsigned char buf[70];

    if (pcl_read_all(reader,buf,arg)) return -1;

    if (buf[0] != 0 || buf[1] != 68 || buf[2] != 20) {
        hpgs_log(hpgs_i18n("PCL Font Header command with unknown font format.\n"));
        return 0;
    }

    int x_res =  (((int)buf[64]) << 8) | ((int)buf[65]);
    int y_res =  (((int)buf[66]) << 8) | ((int)buf[67]);
    double baseline_position = ((((int)buf[6]) << 8) | ((int)buf[7])) * 72.0 / ((double)y_res);

    // 72/64 = 1.125
    double pitch = ((((int)buf[16])<<16)|(((int)buf[17])<<8)|((int)buf[40])) * 1.125 /((double)x_res);
    double height = ((((int)buf[18])<<16)|(((int)buf[19])<<8)|((int)buf[41])) * 1.125 /((double)y_res);

    reader->pcl_current_bitmap_font =
            hpgs_reader_pcl_bitmap_font_new(reader,reader->pcl_current_bitmap_font_id,
            pitch,height,baseline_position,x_res,y_res);

    return reader->pcl_current_bitmap_font == NULL ? -1 : 0;
}

int hpgs_reader_pcl_cmd_select_font_id       (hpgs_reader *reader, int arg) {

    size_t i = hpgs_reader_pcl_bitmap_font_lower_bound(reader,arg);

     if (i < reader->pcl_bitmap_fonts_size &&
             reader->pcl_bitmap_fonts[i]->font_id == arg) {
         reader->pcl_primary_bitmap_font = reader->pcl_bitmap_fonts[i];
     }
     else {
         reader->pcl_primary_bitmap_font = NULL;
     }
     return 0;
}

static int mk_signed_int(unsigned char msb, unsigned char lsb) {

    int ret = ((int)msb << 8) | (int)lsb;

    if (msb &0x80)
        ret |= ~0xffff;

    return ret;
}

int hpgs_reader_pcl_cmd_character_definition (hpgs_reader *reader, int arg) {

    if (reader->pcl_current_character_code < 0 || reader->pcl_current_character_code >= 256) {

        return hpgs_set_error(hpgs_i18n("PCL Character Definition command with invalid character code %d."),reader->pcl_current_character_code);
    }

    if (arg < 4 || arg >= 32768) {

        return hpgs_set_error(hpgs_i18n("PCL Character Definition command with invalid length %d."),arg);
    }

    unsigned char *buf = (unsigned char*)hpgs_alloca(arg);

    if (pcl_read_all(reader,buf,arg)) return -1;

    if (buf[0] != 4 || buf[1] != 0 || buf[2] != 14) {
        hpgs_log(hpgs_i18n("PCL Character Definition command with unknown format.\n"));
        return 0;
    }

    if (reader->pcl_current_bitmap_font == NULL) {
        hpgs_log(hpgs_i18n("PCL Character Definition while no font has been selected.\n"));
        return 0;
    }

    /*

     Byte | 15 (MSB)           8 | 7             (LSB) 0
     ---------------------------------------------------
        0 | Format (4)           | Continuation (0)
        2 | Descriptor Size (14) | Class (1)
        4 | Orientation          | Reserved (0)
        6 | Left Offset
        8 | Top Offset
       10 | Character Width
       12 | Character Height
       14 | Delta X
       16 | Raster Character Data: (in bytes)

     */

    int cls         = buf[3];

    if (cls != 1 && cls != 2) {
        hpgs_log(hpgs_i18n("PCL Character Definition command with unknown class %d.\n"),cls);
        return 0;
    }

    int width       = (((int)buf[10]) << 8) | ((int)buf[11]);
    int height      = (((int)buf[12]) << 8) | ((int)buf[13]);

    if (width <= 0 || width >= 1024 || height <= 0 || height >= 1024) {

         return hpgs_set_error(hpgs_i18n("PCL bitmap character with invalid dimensions %dx%d."),width,height);
    }

    hpgs_reader_pcl_bitmap_character *ch =
            (hpgs_reader_pcl_bitmap_character*)malloc(sizeof(hpgs_reader_pcl_bitmap_character));

    if (!ch)
        return hpgs_set_error(hpgs_i18n("Out of memory allocating new PCL bitmap character."));

    ch->width       = width;
    ch->height      = height;
    ch->top_offset  = mk_signed_int(buf[8],buf[9]);
    ch->left_offset = mk_signed_int(buf[6],buf[7]);
    ch->delta_x     = mk_signed_int(buf[14],buf[15]);
    ch->orientation = buf[4];
    ch->image = NULL;

    // interpret image data only for hte final rendering passto a non-plotsize device.
    if ((hpgs_device_capabilities(reader->device) & HPGS_DEVICE_CAP_PLOTSIZE) == 0) {

        ch->image =
                (hpgs_image *)hpgs_new_png_image(width,height,
                        8,HPGS_TRUE,HPGS_FALSE);

        hpgs_palette_color palette[2] = {
                { 255,255,255 },
                {   0,  0,  0 }
        };

        if (!ch->image || hpgs_image_set_palette(ch->image,palette,2) < 0) {

            free(ch);
            return hpgs_set_error(hpgs_i18n("Error setting up PCL bitmap character image."));
        }

        hpgs_paint_color c = { 0,0,0,0 };

        if (cls == 1) {

            // uncompressed.
            int i = 16,row,col;
            for (row = 0;row<height&& i < arg;++row) {

                int mask = 0;
                int v=0;

                for (col = 0;col<width;++col) {

                    if (mask == 0) {
                        if (i >= arg) break;
                        mask = 0x80;
                        v = buf[i++];
                    }

                    c.index = (v & mask) != 0;
                    hpgs_image_put_pixel(ch->image,col,row,&c,1.0);

                    mask >>= 1;
                }
            }
        }
        else {

            // compressed.
            int i = 16,row,col;
            for (row = 0;row<height&& i < arg;++row) {

                // line repetition
                int nrep = buf[i++];

                if (row+nrep >= height) {
                    nrep = height - 1 - row;
                }

                // start with white
                c.index = 1;

                for (col = 0;col<width;) {

                    int v = buf[i++];

                    if (v > 0) {

                        int irep;

                        if (col+v > width)
                            v = width - col;

                        // line repetition
                        for (irep = 0; irep <= nrep; ++irep)
                            hpgs_image_put_chunk(ch->image,col,col+v,row+irep,&c);
                    }

                    if (i >= arg) break;
                    col += v;
                    c.index = !c.index;
                }
                row += nrep;
            }
        }


        // save inline images, if specified.
        if (reader->png_dump_filename) {
            int l = strlen(reader->png_dump_filename)+16;
            char *fn = hpgs_alloca(l);

            ++reader->png_dump_count;

            size_t pos;
            hpgs_istream_tell(reader->in,&pos);

            snprintf(fn,l,"%s%04d.png",reader->png_dump_filename,reader->png_dump_count);

            if (reader->verbosity)
                hpgs_log(hpgs_i18n("Dumping PCL character image at file position %lu to file <%s>.\n"),(unsigned long)pos,fn);

#ifdef DEBUG_PCL_CHARACTERS
            fprintf(stderr,"w,h,left_off,top_off,dx,orientation = %d,%d,%d,%d,%d,%d.\n",
                    ch->width,ch->height,ch->left_offset,ch->top_offset,ch->delta_x,ch->orientation);
#endif
            int ret = hpgs_image_write(ch->image,fn);

            if (ret) { hpgs_image_destroy(ch->image); free(ch); return -1; }
        }
    }

    reader->pcl_current_bitmap_font->characters[reader->pcl_current_character_code] = ch;

    return 0;
}

int hpgs_reader_pcl_bitmap_character_render(hpgs_reader *reader, const hpgs_reader_pcl_bitmap_font *font, int charcode) {

    const hpgs_reader_pcl_bitmap_character *ch = font->characters[charcode];

    if (ch == NULL) return 0;

    double xfac = 72.0 / (double)font->x_resolution;
    double yfac = 72.0 / (double)font->y_resolution;

    if ((hpgs_device_capabilities(reader->device) & HPGS_DEVICE_CAP_NULLIMAGE) ||
            ch->image) {

        hpgs_point ll,lr,ur;

        ur.x = reader->pcl_point.x - ch->top_offset * yfac;
        ll.y = reader->pcl_point.y + ch->left_offset * xfac;

        ur.y = ll.y + ch->width * xfac;
        ll.x = ur.x + ch->height * yfac;

        lr.x = ll.x;
        lr.y = ur.y;

        hpgs_point ll_p;
        hpgs_point lr_p;
        hpgs_point ur_p;

        hpgs_matrix_xform(&ll_p,&reader->page_matrix,&ll);
        hpgs_matrix_xform(&lr_p,&reader->page_matrix,&lr);
        hpgs_matrix_xform(&ur_p,&reader->page_matrix,&ur);

   #ifdef HPGS_PCL_DEBUG
        hpgs_log("ll_p,lr_p,ur_p=(%lg,%lg),(%lg,%lg),(%lg,%lg).\n",
                ll_p.x,ll_p.y,lr_p.x,lr_p.y,ur_p.x,ur_p.y);
   #endif

        hpgs_color patcol;

        double fac =
                reader->pcl_pattern ? (reader->pcl_pattern == 1 ? 0.0 : 0.5) : 1.0;

        patcol.r = (1.0-fac) + fac * (1.0/256.0) * reader->pcl_foreground_color.r;
        patcol.g = (1.0-fac) + fac * (1.0/256.0) * reader->pcl_foreground_color.g;
        patcol.b = (1.0-fac) + fac * (1.0/256.0) * reader->pcl_foreground_color.b;

        if (hpgs_setpatcol(reader->device,&patcol))
            return -1;

        // draw image
        if (hpgs_drawimage(reader->device,ch->image,&ll_p,&lr_p,&ur_p))
            return -1;

        if (hpgs_setpatcol(reader->device,&reader->pen_colors[reader->current_pen]))
            return -1;
    }

    switch (ch->orientation == 0) {
    default:
        reader->pcl_point.x += xfac * ch->delta_x;
        break;
    case 1:
        reader->pcl_point.y += yfac * ch->delta_x;
        break;
    case 2:
        reader->pcl_point.x -= xfac * ch->delta_x;
        break;
    case 3:
        reader->pcl_point.y -= yfac * ch->delta_x;
        break;
    }
    return 0;
}
